#ifndef VNET_YARP_PROXY_HH_
#define VNET_YARP_PROXY_HH_

#include <string>
#include <yarp/os/all.h>

#include <vnet_filter.hh>
#include <vnet_table.hh>

namespace vnet {
	namespace yarp {
		class BufferReader : public ::yarp::os::PortReader {
			private:
				char * data_;
				int len_data_;
				int real_len_;

				BufferReader(const BufferReader&);
				BufferReader& operator=(const BufferReader&);

			public:
				BufferReader() : data_(0), len_data_(0), real_len_(0)
				{}

				bool read(::yarp::os::ConnectionReader& reader) {
					int data_size = reader.getSize();
					if (data_size > real_len_) {
						delete [] data_;
						data_ = new char[data_size];
						real_len_ = data_size;
					}
					len_data_ = data_size;
					return reader.expectBlock(data_, len_data_);
				}

				~BufferReader() { delete [] data_; }

				const char* data() const { return data_; }
				const int len() const { return len_data_; }
		};

		class BufferWriter : public ::yarp::os::PortWriter {
			private:
				const char* data;
				const int len;

			public:
				BufferWriter(const char* data, int len) :
					data(data), len(len)
				{}

				bool write(::yarp::os::ConnectionWriter& writer) {
					writer.appendExternalBlock(data, len);
					return !writer.isError();
				}
		};

		class ReaderProxy {
			private:
				::yarp::os::Port port;
				BufferReader reader;
				std::pair<std::string, std::string> envelop_;

			public:
				ReaderProxy(const std::string& uri) : port() {
					port.open(uri.c_str());
				}

				bool read() {
					bool res = port.read(reader, true);

					::yarp::os::Bottle b;
					port.getEnvelope(b);

					envelop_.first = b.get(0).asString().c_str();
					envelop_.second = b.get(1).asString().c_str();
					return res;
				}

				void reply(bool success) {
					::yarp::os::Bottle b;
					b.addInt(success);

					port.reply(b);
				}

				const char* data() { return reader.data(); }
				const int len() { return reader.len(); }
				const std::pair<std::string, std::string>& envelop() { return envelop_; }
		};

		class WriterProxy {
			private:
				::yarp::os::Port port;
				std::string local_uri;
				std::string current_dst;

			public:
				WriterProxy(const std::string& local_uri) : 
					port(), local_uri(local_uri), current_dst("")
				{
					port.open(local_uri.c_str());
				}

				bool write(const std::string& dst, const char* data, int len) 
				{
					if (current_dst != dst) {
						port.close();
						port.open(local_uri.c_str());
						port.addOutput(dst.c_str());
						current_dst = dst;
					}
					BufferWriter writer(data, len);
					return port.write(writer);
				}
		};


		class vnetProxy {
			private:
				ReaderProxy reader;
				WriterProxy writer;
				std::string current_dst;

				table& m_table;

				std::string extract_robot_name(const std::string& s)
				{
					std::size_t first, second;
					first = s.find_first_of("/");
					assert(first != std::string::npos);
					second = s.find_first_of("/", first + 1);
					assert(second != std::string::npos);
					return s.substr(first+1, second-first-1);
				}

				std::pair<std::string, std::string>
				get_robot_pair(const std::pair<std::string, std::string>& p)
				{
					std::string fst = extract_robot_name(p.first);
					std::string snd = extract_robot_name(p.second);
					return std::make_pair(fst, snd);
				}

			public:
				vnetProxy(table& m_table):
					reader("/vnet/in"), writer("/vnet/out"), m_table(m_table)
				{}

				void run()
				{
					reader.read();
					std::pair<std::string, std::string> env = reader.envelop();
					std::pair<std::string, std::string> robots = get_robot_pair(env);
					boost::lock_guard<table> lk(m_table);
					const table::filter_array& arr = m_table.get(robots);
					table::filter_array::const_iterator it;
					bool res = true;
					for (it = arr.begin(); it != arr.end() and res; ++it)
					{
						res = res and (*it)->filter(robots);
					}
					if (res) {
						res = writer.write(env.second, reader.data(), reader.len());
					}

					reader.reply(res);
				}
		};
	}
}


#endif /* VNET_YARP_PROXY_HH_ */
